;******************************************************************************
; file:      stdout.asm
; content:   sf20b standard out routines
; author:    Martin Raubuch
; created:   22Dec2014
; modified:  22Dec2014
; Property of RACORS GmbH, Ginsterweg 1a, 85598 Baldham, Germany
;******************************************************************************
; ABI Conventions, allocation order
; R8           used as stack/frame pointer
; RF-RC        pointer argument passing, may be destroyed by called function
; R0-R3        data argument passing, may be destroyed by called function
; R0           return value
; R4-R7/R9-RB  if modified must be saved/restored by called function	
;******************************************************************************

;**************************** puts (put string) *******************************
; writes the specified 0-terminated string to the standard output (sim. window)
; arg RF = string pointer
; returns nothing
;******************************************************************************
puts:
   ldbt   (RF)+,"01"		; R0/R1 = next 2 bytes of string
   comp   0,R0			; R0 == 0 ? termination byte
   brzr   1f			; if yes continue at 1f
   comp   0,R1			; R1 == 0 ? termination byte
   stbt   R0,StdOut		; if not write to std-out
   brzr   1f			; if yes continue at 1f
   stbt   R1,StdOut		; if not write to std-out
   bral   puts			; loop back to puts
1: rtsr				; quit
;************************* putchar (put character) ****************************
; writes the specified character to the standard output (simulation window)
; arg R0 = character
; returns nothing
;******************************************************************************
putchar:
   stbt   R0,StdOut		; write to std-out
   rtsr				; quit
;********************** print simulation cycle number *************************
; prints the current simulation cycle number as "@nnnnnnnnn  "
; arg none
; returns nothing
;******************************************************************************
simcycle:
   stbt   R0,SimCyc		; write to &SimCyc, trigger, data don't care
   rtsr				; quit
