/******************************************************************************
file:      sf20bu.v
content:   RACORS sf20bu 16-bit microprocessor, synthesizable Verilog RTL
           member of the sf20 family of 16-bit microprocessors
           b (base) ISA (Instruction Set Architecture)
           u (ultralight) IMA (Implementation Architecture)
created:   22Dec2014
*******************************************************************************
  Property of RACORS GmbH, Ginsterweg 1a, 85598 Baldham, Germany
  Licensing terms are specified in a separate document
******************************************************************************/
module sf20bu(CLK,RST,II,IRDY,IRN,IRQ,DBGI,STRQ,INJI,DI,DRDY,IA,IFT,IFNS,
              IACK,DBGO,STPD,DA,DO,DBS,DWE);
  input         CLK;
  input         RST;
  input  [19:0] II;
  input         IRDY;
  input  [ 3:0] IRN;
  input         IRQ;
  input  [19:0] DBGI;
  input         STRQ;
  input         INJI;
  input  [15:0] DI;
  input         DRDY;
  output [15:0] IA;
  output        IFT;
  output        IFNS;
  output        IACK;
  output [15:0] DBGO;
  output        STPD;
  output [15:0] DA;
  output [15:0] DO;
  output [ 1:0] DBS;
  output        DWE;
  wire   [15:0] CTU_IA;
  wire          CTU_IFT;
  wire          CTU_IFNS;
  wire          CTU_IACK;
  wire   [15:0] CTU_DBGO;
  wire          CTU_STPD;
  wire          CTU_IRST;
  wire   [ 3:0] CTU_CC;
  wire   [ 3:0] CTU_CCS;
  wire   [ 2:0] CTU_CSF;
  wire   [ 3:0] CTU_LIRN;
  wire          CTU_IVFT;
  wire   [ 4:0] CTU_RRS;
  wire   [19:0] CTU_DBGIR;
  wire   [15:0] CTU_PC;
  wire   [ 1:0] CTU_IRQST;
  wire   [19:0] CTU_IW;
  wire   [19:0] CTU_DIW;
  wire          CTU_DIWV;
  wire   [ 2:0] CTU_DAS0CT;
  wire   [ 2:0] CTU_DAS1CT;
  wire   [ 4:0] CTU_DACOCT;
  wire   [ 4:0] CTU_DALUCT;
  wire   [ 2:0] CTU_DLSCT;
  wire          CTU_DLSASL;
  wire   [ 4:0] CTU_DRRS;
  wire          CTU_AEXRD;
  wire          CTU_ALUWP;
  wire   [ 1:0] CTU_RFE0;
  wire   [ 1:0] CTU_RFE1;
  wire   [ 4:0] CTU_RWS;
  wire   [ 4:0] CTU_RBWA;
  wire          CTU_RWE;
  wire   [ 1:0] CTU_DCCOPC;
  wire   [ 1:0] CTU_CCOP;
  wire          CTU_FLIV;
  wire   [ 1:0] CTU_IBFL;
  wire          CTU_FEXR;
  wire   [ 1:0] CTU_IPH;
  wire   [ 1:0] CTU_NIPH;
  wire          CTU_STALL;
  wire   [15:0] DPU_DO;
  wire   [ 3:0] DPU_CFO;
  wire   [15:0] DPU_OR0;
  wire   [15:0] DPU_MDA;
  wire   [15:0] DPU_MDO;
  wire   [ 1:0] DPU_MDBS;
  wire          DPU_MDWE;
  wire          DPU_MBFL;
  wire   [ 4:0] DPU_MBDR;
  wire   [ 4:0] DPU_MADR;
  wire          DPU_MAWE;
  wire   [15:0] RF_RP;
  wire   [15:0] RF_RDTW;
  wire   [15:0] RFB_DO;
  assign IA[15:0]   = CTU_IA[15:0];
  assign IFT        = CTU_IFT;
  assign IFNS       = CTU_IFNS;
  assign IACK       = CTU_IACK;
  assign DBGO[15:0] = CTU_DBGO[15:0];
  assign STPD       = CTU_STPD;
  assign DA[15:0]   = DPU_MDA[15:0];
  assign DO[15:0]   = DPU_MDO[15:0];
  assign DBS[ 1:0]  = DPU_MDBS[1:0];
  assign DWE        = DPU_MDWE;
  sf20buCtu CTU (
  .CLK     (CLK),
  .RST     (RST),
  .IRDY    (IRDY),
  .II      (II[19:0]),
  .IRN     (IRN[3:0]),
  .IRQ     (IRQ),
  .DBGI    (DBGI[19:0]),
  .STRQ    (STRQ),
  .INJI    (INJI),
  .MDI     (DI[15:0]),
  .MDRDY   (DRDY),
  .RBDI    (RFB_DO[15:0]),
  .CFI     (DPU_CFO[3:0]),
  .DS0DI   (DPU_OR0[15:0]),
  .MBFL    (DPU_MBFL),
  .MBDR    (DPU_MBDR[4:0]),
  .MBWE    (DPU_MDWE),
  .MADR    (DPU_MADR[4:0]),
  .MAWE    (DPU_MAWE),
  .IA      (CTU_IA[15:0]),
  .IFT     (CTU_IFT),
  .IFNS    (CTU_IFNS),
  .IACK    (CTU_IACK),
  .DBGO    (CTU_DBGO[15:0]),
  .STPD    (CTU_STPD),
  .IRST    (CTU_IRST),
  .CC      (CTU_CC[3:0]),
  .CCS     (CTU_CCS[3:0]),
  .CSF     (CTU_CSF[2:0]),
  .LIRN    (CTU_LIRN[3:0]),
  .IVFT    (CTU_IVFT),
  .RRS     (CTU_RRS[4:0]),
  .DBGIR   (CTU_DBGIR[19:0]),
  .PC      (CTU_PC[15:0]),
  .IRQST   (CTU_IRQST[1:0]),
  .IW      (CTU_IW[19:0]),
  .DIW     (CTU_DIW[19:0]),
  .DIWV    (CTU_DIWV),
  .DAS0CT  (CTU_DAS0CT[2:0]),
  .DAS1CT  (CTU_DAS1CT[2:0]),
  .DACOCT  (CTU_DACOCT[4:0]),
  .DALUCT  (CTU_DALUCT[4:0]),
  .DLSCT   (CTU_DLSCT[2:0]),
  .DLSASL  (CTU_DLSASL),
  .DRRS    (CTU_DRRS[4:0]),
  .AEXRD   (CTU_AEXRD),
  .ALUWP   (CTU_ALUWP),
  .RFE0    (CTU_RFE0[1:0]),
  .RFE1    (CTU_RFE1[1:0]),
  .RWS     (CTU_RWS[4:0]),
  .RBWA    (CTU_RBWA[4:0]),
  .RWE     (CTU_RWE),
  .DCCOPC  (CTU_DCCOPC[1:0]),
  .CCOP    (CTU_CCOP[1:0]),
  .FLIV    (CTU_FLIV),
  .IBFL    (CTU_IBFL[1:0]),
  .FEXR    (CTU_FEXR),
  .IPH     (CTU_IPH[1:0]),
  .NIPH    (CTU_NIPH[1:0]),
  .STALL   (CTU_STALL)
  );
  sf20buDpu DPU (
  .CLK     (CLK),
  .RST     (CTU_IRST),
  .MDI     (DI[15:0]),
  .MDRDY   (DRDY),
  .RSD     (RF_RP[15:0]),
  .RFE0    (CTU_RFE0[1:0]),
  .RFE1    (CTU_RFE1[1:0]),
  .IW      (CTU_DIW[19:0]),
  .DBGI    (CTU_DBGIR[19:0]),
  .CC      (CTU_CC[3:0]),
  .EXRD    (CTU_AEXRD),
  .ALUWP   (CTU_ALUWP),
  .S0CT    (CTU_DAS0CT[2:0]),
  .S1CT    (CTU_DAS1CT[2:0]),
  .COCT    (CTU_DACOCT[4:0]),
  .DCCOPC  (CTU_DCCOPC[1:0]),
  .CCOP    (CTU_CCOP[1:0]),
  .DRRS    (CTU_DRRS[4:0]),
  .RWS     (CTU_RWS[4:0]),
  .RWE     (CTU_RWE),
  .LSCT    (CTU_DLSCT[2:0]),
  .LSASL   (CTU_DLSASL),
  .IVFT    (CTU_IVFT),
  .LIRN    (CTU_LIRN[3:0]),
  .PC      (CTU_PC[15:0]),
  .IRQST   (CTU_IRQST[1:0]),
  .OPCT    (CTU_DALUCT[4:0]),
  .DO      (DPU_DO[15:0]),
  .CFO     (DPU_CFO[3:0]),
  .OR0     (DPU_OR0[15:0]),
  .MDA     (DPU_MDA[15:0]),
  .MDO     (DPU_MDO[15:0]),
  .MDBS    (DPU_MDBS[1:0]),
  .MDWE    (DPU_MDWE),
  .MBFL    (DPU_MBFL),
  .MBDR    (DPU_MBDR[4:0]),
  .MADR    (DPU_MADR[4:0]),
  .MAWE    (DPU_MAWE)
  );
  sf20buRf RF (
  .CLK   (CLK),
  .RST   (CTU_IRST),
  .DRRS  (CTU_DRRS[4:0]),
  .CC    (CTU_CC[3:0]),
  .CCS   (CTU_CCS[3:0]),
  .CSF   (CTU_CSF[2:0]),
  .RBDO  (RFB_DO[15:0]),
  .RBDI  (DPU_DO[15:0]),
  .RBWA  (CTU_RBWA[4:0]),
  .RBWE  (CTU_RWE),
  .DI    (DPU_DO[15:0]),
  .RP    (RF_RP[15:0]),
  .RDTW  (RF_RDTW[15:0])
  );
  sf20buRfb RFB (
  .CLK  (CLK),
  .RAD  (CTU_RRS[4:0]),
  .DI   (RF_RDTW[15:0]),
  .WAD  (CTU_RBWA[4:0]),
  .WE   (CTU_RWE),
  .DO   (RFB_DO[15:0])
  );
endmodule
module sf20buCtu(CLK,RST,II,IRDY,IRN,IRQ,DBGI,STRQ,INJI,MDI,MDRDY,CFI,DS0DI,
                 RBDI,MBFL,MBDR,MBWE,MADR,MAWE,IA,IFT,IFNS,IACK,DBGO,STPD,
                 IRST,CC,CCS,CSF,LIRN,IVFT,RRS,DBGIR,PC,IRQST,IW,DIW,DIWV,
                 DAS0CT,DAS1CT,DACOCT,DALUCT,DLSCT,DLSASL,DRRS,AEXRD,ALUWP,
                 RFE0,RFE1,RWS,RBWA,RWE,DCCOPC,CCOP,FLIV,IBFL,FEXR,IPH,NIPH,
                 STALL);
  input         CLK;
  input         RST;
  input  [19:0] II;
  input         IRDY;
  input  [ 3:0] IRN;
  input         IRQ;
  input  [19:0] DBGI;
  input         STRQ;
  input         INJI;
  input  [15:0] MDI;
  input         MDRDY;
  input  [ 3:0] CFI;
  input  [15:0] DS0DI;
  input  [15:0] RBDI;
  input         MBFL;
  input  [ 4:0] MBDR;
  input         MBWE;
  input  [ 4:0] MADR;
  input         MAWE;
  output [15:0] IA;
  output        IFT;
  output        IFNS;
  output        IACK;
  output [15:0] DBGO;
  output        STPD;
  output        IRST;
  output [ 3:0] CC;
  output [ 3:0] CCS;
  output [ 2:0] CSF;
  output [ 3:0] LIRN;
  output        IVFT;
  output [ 4:0] RRS;
  output [19:0] DBGIR;
  output [15:0] PC;
  output [ 1:0] IRQST;
  output [19:0] IW;
  output [19:0] DIW;
  output        DIWV;
  output [ 2:0] DAS0CT;
  output [ 2:0] DAS1CT;
  output [ 4:0] DACOCT;
  output [ 4:0] DALUCT;
  output [ 2:0] DLSCT;
  output        DLSASL;
  output [ 4:0] DRRS;
  output        AEXRD;
  output        ALUWP;
  output [ 1:0] RFE0;
  output [ 1:0] RFE1;
  output [ 4:0] RWS;
  output [ 4:0] RBWA;
  output        RWE;
  output [ 1:0] DCCOPC;
  output [ 1:0] CCOP;
  output        FLIV;
  output [ 1:0] IBFL;
  output        FEXR;
  output [ 1:0] IPH;
  output [ 1:0] NIPH;
  output        STALL;
  reg    [15:0] IA;
  reg           IFT;
  reg           IFNS;
  reg           IACK;
  reg    [15:0] DBGO;
  reg           STPD;
  reg           IRST;
  reg    [ 3:0] CC;
  reg    [ 3:0] CCS;
  reg    [ 2:0] CSF;
  reg    [ 3:0] LIRN;
  reg           IVFT;
  reg    [ 4:0] RRS;
  reg    [19:0] DBGIR;
  reg    [15:0] PC;
  reg    [ 1:0] IRQST;
  reg    [19:0] IW;
  reg    [19:0] DIW;
  reg           DIWV;
  reg    [ 2:0] DAS0CT;
  reg    [ 2:0] DAS1CT;
  reg    [ 4:0] DACOCT;
  reg    [ 4:0] DALUCT;
  reg    [ 2:0] DLSCT;
  reg           DLSASL;
  reg    [ 4:0] DRRS;
  reg           AEXRD;
  reg           ALUWP;
  reg    [ 1:0] RFE0;
  reg    [ 1:0] RFE1;
  reg    [ 4:0] RWS;
  reg    [ 4:0] RBWA;
  reg           RWE;
  reg    [ 1:0] DCCOPC;
  reg    [ 1:0] CCOP;
  reg           FLIV;
  reg    [ 1:0] IBFL;
  reg           FEXR;
  reg    [ 1:0] IPH;
  reg    [ 1:0] NIPH;
  reg           STALL;
  reg    [ 1:0] RSYN;
  reg    [15:0] MAIA;
  reg    [19:0] IB0;
  reg    [19:0] IB1;
  reg           MAST;
  reg           LCTST;
  reg           NSIF;
  reg           DSOIR;
  reg           LSTRQ;
  reg           LINJI;
  reg           LDBRQ;
  reg           INDB;
  reg           RWES;
  reg    [ 1:0] CCOPS;
  reg           LCTDS;
  reg           DLDSTM;
  reg           DRRE;
  reg    [ 4:0] DRWE;
  reg           DDBGC;
  reg           DBGWE;
  reg    [ 3:0] SHCT;
  reg    [15:0] BTIA;
  reg    [15:0] BOFS;
  reg    [ 2:0] IACT;
  reg    [ 2:0] FLCT;
  reg    [ 1:0] CCSCT;
  reg           STPCT;
  reg           BCDT;
  reg    [ 1:0] PCCT;
  reg           IIV;
  reg           IWCS;
  reg    [ 2:0] QDIW;
  reg           CCPO;
  reg           LCPO;
  reg           TAPO;
  reg           SAPO;
  reg           SOIR;
  reg           RFST;
  reg    [ 2:0] NCCSF;
  reg           NCMAST;
  reg           NCNSIF;
  reg    [ 1:0] NCIRQST;
  reg    [ 1:0] NCIBFL;
  reg           NCLDBRQ;
  reg           SQFFI;
  reg    [ 1:0] LCTLD;
  reg           RFNXI;
  reg           STRR;
  reg           STRW;
  reg    [ 4:0] SMRS;
  reg    [ 4:0] LMRS;
  reg    [19:0] RFCM;
  reg    [ 5:0] RRSC;
  reg    [ 4:0] RWEC;
  reg           DBGC;
  reg    [ 2:0] AS0CT;
  reg    [ 2:0] AS1CT;
  reg    [ 4:0] ACOCT;
  reg    [ 4:0] ALUCT;
  reg    [ 1:0] CCOPC;
  reg    [ 2:0] LSCT;
  reg           LSASL;
  reg           LDSTM;
  reg           LCTD;
  reg    [ 4:0] NCRWS;
  reg           NCRWES;
  reg    [ 3:0] NCSHCT;
  reg               c;     
  reg               o;     
  reg               z;     
  reg               n;     
  task idec_jpsr_ia16; begin
    IACT[0]  = 1;
    NCNSIF   = 1;
    RWEC     = 5'h1c;                   
    AS0CT[1] = 1;                       
    AS1CT[2] = 1;                       
  end endtask
  task idec_jump_ta; begin
    if(IPH == 0) begin
      if(~TAPO) begin
        NIPH     = 1;                   
        AS0CT[0] = 1;                   
        RRSC     = 6'h3e;               
      end
      FEXR = 0;
    end
    if(IPH == 1) begin
      NIPH    = 0;                      
      IACT[2] = 1;
      NCNSIF  = 1;
      if(IW[4]) begin                   
        RWEC     = 5'h1c;               
        AS0CT[1] = 1;                   
        AS1CT[2] = 1;                   
      end
    end
  end endtask
  task idec_brlc; begin
    FEXR = ~LCPO;
    if(FEXR) begin
      if(LCTST) begin
        IACT[1] = 1;
        NCNSIF  = 1;
      end
      else
        SQFFI  = 1;
      RWEC     = 5'h12;                 
      RRSC     = 6'h32;                 
      AS0CT[0] = 1;                     
      ACOCT[2:0] = 5;                   
      AS1CT[1] = 1;                     
      ALUCT[0] = 1;                     
    end
  end endtask
  task idec_bral; begin
    IACT[1] = 1;
    NCNSIF  = 1;
  end endtask
  task idec_return; begin
    if(IPH == 0) begin
      FEXR = 0;                         
      if(IW[4] | ~SAPO) begin
        NIPH     = 1;                   
        AS0CT[0] = 1;                   
        RRSC     = {5'h1e,IW[4]};       
      end
    end
    if(IPH == 1) begin
      NIPH    = 0;                      
      IACT[2] = 1;                      
      NCNSIF  = 1;                      
      if(IW[4])                         
        CCSCT[1] = 1;                   
    end
  end endtask
  task idec_cdbranch; begin
    FEXR = ~IRQST[1] & ~CCPO;
    if(FEXR & BCDT) begin
      IACT[1] = 1;
      NCNSIF  = 1;
    end
    else
      SQFFI = 1;
  end endtask
  task idec_stop; begin
    STPCT = 1;
    FEXR  = IRQST == 0;
  end endtask
  task idec_con8; begin
    RWEC[0]  = 1;                       
    CCOPC[0] = IW[10:9] == 0;           
    CCOPC[0] = CCOPC[0] | (IW[10:8]==5); 
    AS0CT[1] = 1;                       
    RRSC[2]  = 1;                       
    AS1CT[0] = 1;                       
    ACOCT    = IW[10:8] == 3 ? 5'h10:5'h8; 
    case(IW[10:8])
     1,
     3: ALUCT = 5'h01;                  
     4: ALUCT = 5'h08;                  
     5: ALUCT = 5'h11;                  
     6: ALUCT = 5'h12;                  
     7: ALUCT = 5'h13;                  
     default: ;
    endcase
  end endtask
  task idec_con10; begin
    AS0CT[1]   = 1;                     
    ACOCT[3:0] = 9;                     
    if(IW[3:2] == 1) begin              
      RWEC     = 1;                     
      ACOCT[2] = 1;                     
      if(IW[8]) begin                   
        RRSC     = 6'h28;               
        AS1CT[0] = 1;                   
        ALUCT[0] = 1;                   
      end
      else
        ALUCT[2] = 1;                   
    end
    if(IW[3:2] == 2) begin              
      CCOPC[0] = 1;                     
      ACOCT[2] = 1;                     
      RRSC[2]  = 1;                     
      AS1CT[0] = 1;                     
    end
    if(IW[3:2] == 3) begin              
      ALUCT[2] = 1;                     
      RWEC     = {1'b1,IW[7:4]};        
    end
  end endtask
  task idec_compreg; begin
    if(IPH == 1) begin
      NIPH     = 0;                     
      RRSC[2]  = 1;                     
      AS1CT[0] = 1;                     
      CCOPC[0] = 1;                     
      CCOPC[1] = IW[12];                
    end
    else begin
      NIPH     = 1;                     
      RRSC[1]  = 1;                     
      AS0CT[0] = 1;                     
    end
  end endtask
  task idec_dual; begin
    RWEC[0]  = IW[15:13] != 5;          
    CCOPC    = IW[15] ? 2'd3:2'd0;      
    RRSC[1]  = 1;                       
    if(IW[15]) begin                    
      AS0CT[1] = 1;                     
      AS1CT[0] = 1;                     
    end
    else begin
      AS0CT[0] = 1;                     
      AS1CT[1] = 1;                     
    end
    ACOCT[2:0] = IW[14] ? 3'd5 : 3'd0;  
    case(IW[15:12])
      0: ALUCT = 5'h04;                 
      3: ALUCT = 5'h02;                 
      4: ALUCT = 5'h13;                 
      5: ALUCT = 5'h06;                 
      6: ALUCT = 5'h07;                 
      7: ALUCT = 5'h03;                 
      9: ALUCT = 5'h01;                 
      default: ;
    endcase
  end endtask
  task idec_shfbit; begin
    ALUCT = {1'b1,IW[18]?2'd1:2'd2,IW[17:16]}; 
    if(IPH == 0)
      if(~IW[2]) begin                  
        NIPH    = 1;                    
        RRSC[0] = 1;                    
      end
      else begin
        RRSC[1]  = 1;                   
        AS1CT[0] = 1;                   
        if(~IW[18] & (IW[15:13]!=0))    
          NIPH = 2;                     
        else if(IW[18:16] == 5)         
          CCOPC[0] = 1;                 
        else
          RWEC[0]  = 1;                 
      end
    if(IPH == 1) begin
      if(IW[18]) begin                  
        NIPH     = 0;                   
        RRSC[1]  = 1;                   
        AS1CT[0] = 1;                   
        if(IW[17:16] == 1)              
          CCOPC[0] = 1;                 
        else
          RWEC[0]  = 1;                 
      end
      else
        NIPH = 3;                       
    end
    if((IPH==2) & (SHCT[3:1]==0)) begin
      NIPH    = 0;                      
      RWEC[0] = 1;                      
    end
    if(IPH == 3) begin
      RRSC[1]  = 1;                     
      AS0CT[0] = 1;                     
      AS1CT[0] = 1;                     
      if(SHCT[3:1] == 0) begin
        NIPH    = 0;                    
        RWEC[0] = 1;                    
        if(~SHCT[0])                    
          ALUCT = 4;                    
      end
      else
        NIPH = 2;                       
    end
  end endtask
  task idec_triarn; begin
    if(IW[18])
      ALUCT = {3'd4,IW[17:16]};        
    else
      ALUCT[0] = IW[16];               
    if(IPH == 1) begin
      NIPH    = 0;                     
      RWEC[0] = 1;                     
      if(~IW[18])
        CCOPC = {IW[17],1'b1};         
      if(IW[17:16] == 1)
        CCOPC[0] = 1;                  
      RRSC[1]  = 1;                    
      AS1CT[0] = 1;                    
    end
    else begin
      NIPH     = 1;                    
      RRSC[0]  = 1;                    
      AS0CT[0] = 1;                    
    end
  end endtask
  task idec_mult; begin
    ALUCT[3:0] = {2'd2,IW[17:16]};     
    if(IPH == 1) begin
      NIPH     = 0;                    
      RWEC[0]  = 1;                    
      RRSC[1]  = 1;                    
      AS1CT[0] = 1;                    
    end
    else begin
      NIPH     = 1;                    
      RRSC[0]  = 1;                    
      AS0CT[0] = 1;                    
    end
  end endtask
  task idec_ldst; begin
    LSCT = IW[3:2];                    
    ALUCT[0] = 1;                      
    if(IPH == 0) begin
      NIPH = 1;                        
      if(IW[0]) begin                  
        RRSC[3]    = 1;                
        AS1CT[0]   = 1;                
        ACOCT[3:2] = 3;                
        AS0CT[1]   = 1;                
      end
      else if(~IW[19]) begin           
        ACOCT[3:0] = 4'hb;             
        AS0CT[1] = 1;                  
        ALUCT    = 4;                  
      end
      else begin
        RRSC[3]  = 1;                  
        AS1CT[0] = 1;                  
        case(IW[18:17])
          0: NIPH = 2;                 
          1: case({IW[16],IW[11]})
               0: begin              
                    ACOCT[2:1] = {1'b1,IW[2]}; 
                    AS0CT[1] = 1;    
                  end
               1: begin              
                    ACOCT[2:0] = {1'b1,IW[2],1'b1}; 
                    AS0CT[1] = 1;    
                  end
               default:              
                  NIPH = 2;          
             endcase
          default: begin             
            AS0CT[1] = 1;            
            ACOCT[2:0] = {1'b1,IW[2],IW[11]}; 
            LDSTM = 1;               
	    if((IW[16:12]==0) & (IW[7:4]==0))
              LDSTM = 0;
	    if(~IW[17] & (IW[15:12]==0) & (IW[7:4]==0))
              LDSTM = 0;
	    if((IW[17:16]==0) & (IW[14:12]==0) & (IW[7:4]==0))
              LDSTM = 0;
	    if((IW[17:15]==0) & (IW[13:12]==0) & (IW[7:4]==0))
              LDSTM = 0;
	    if((IW[17:14]==0) & ~IW[12] & (IW[7:4]==0))
              LDSTM = 0;
	    if((IW[17:13]==0) & (IW[7:4]==0))
              LDSTM = 0;
	    if((IW[17:12]==0) & (IW[6:4]==0))
              LDSTM = 0;
	    if((IW[17:12]==0) & ~IW[7] & (IW[5:4]==0))
              LDSTM = 0;
	    if((IW[17:12]==0) & (IW[7:6]==0) & ~IW[4])
              LDSTM = 0;
	    if((IW[17:12]==0) & (IW[7:5]==0))
              LDSTM = 0;
          end
        endcase
      end
    end
    if(IPH == 1) begin
      if(IW[19] & (IW[18:17]!=0))
        RWEC[1] = 1;                   
      if(IW[19:18] == 3) begin         
        RRSC[4] = 1;                   
        if(~IW[11])
          LSASL = 1;                   
      end
      else
        RRSC[2] = 1;                   
      NIPH  = 0;                       
      LSCT[2] = 1;                     
      if(IW[19] & ~IW[18] & IW[17] & ~IW[11]) 
        LSASL = 1;                   
    end
    if(IPH == 2) begin                 
      RRSC     = IW[16] ? 6'h34 : 6'h1; 
      AS0CT[0] = 1;                    
      NIPH     = 1;                    
    end
  end endtask
  task idec_mtfsrdp; begin
    ALUCT[2] = 1;                      
    if(IW[12])                         
      RRSC[1] = 1;                     
    else
      RWEC[0] = 1;                     
    case(IW[13:12])
      0: AS0CT[2] = 1;                 
      1: begin                         
           AS0CT[0] = 1;               
           DBGC     = 1;               
         end
      2: begin                         
           AS0CT[0] = 1;               
           RRSC = {2'd3,IW[11:8]};     
         end
      3: begin                         
           AS0CT[0] = 1;               
           RWEC = {1'b1,IW[7:4]};      
         end
    endcase
  end endtask
  always @(RSYN)
    IRST = RSYN[0];
  always @(IRQST or MBFL)
    IVFT = (IRQST==2) & ~MBFL;
  always @(IBFL or MAST)
    QDIW = IBFL + MAST;
  always @(MAST or IRDY)
    IIV = MAST & IRDY;
  always @(IW or CC) begin
    BCDT = 0;
    c = CC[0];
    o = CC[1];
    z = CC[2];
    n = CC[3];
    case({IW[2],IW[6:5]})
      0: BCDT = ~c;
      1: BCDT = ~o;
      2: BCDT = ~z;
      3: BCDT = ~n;
      4: BCDT =  c | z;
      5: BCDT = (n & ~o) | (~n &  o);
      6: BCDT =  z |  (n & ~o) | (~n &  o);
      default: ;
    endcase
    BCDT = BCDT ^ IW[4];
  end
  always @(RWES or MAWE or MDRDY or RWS or MADR)
    ALUWP = RWES & MAWE & (RWS==MADR);
  always @(RWES or DAS0CT or RWS or DRRS or MAWE or MDRDY or MADR) begin
    RFE0[0] = RWES & DAS0CT[0] & (RWS==DRRS) & (RWS!=5'h1f);
    RFE0[1] = MAWE & MDRDY & (MADR==DRRS);
  end
  always @(RWES or DAS1CT or RWS or DRRS or SHCT or RFE0) begin
    RFE1[0] = RWES & DAS1CT[0] & (RWS==DRRS) & (RWS!=5'h1f);
    RFE1[0] = RFE1[0] | ((SHCT!=0)&~DAS1CT[0]);
    RFE1[1] = RFE0[1];
  end
  always @(DRRE or DIW or RWES or DRRS or RWS or MAWE or MADR or CCOPS or
           MDRDY or MBFL or MBWE or MBDR) begin
    STRR = 0;
    if(DRRE) begin
      if(~DIW[19] & (DIW[3:0]==2))
        STRR = (RWES&(DRRS==RWS)) | (MAWE &(DRRS==MADR));
      STRR = STRR | (CCOPS[0]&(DRRS==5'h10));
      STRR = STRR | (RWES&MAWE&MDRDY&(RWS!=MADR));
      STRR = STRR | (MAWE&~MDRDY&(DRRS==MADR));
      STRR = STRR | (MBFL &~MBWE&(DRRS==MBDR));
    end
  end
  always @(RWES or MAWE or RWS or MADR or MBFL or MBWE or MBDR) begin
    STRW = 0;
    if(RWES)
      STRW = (MAWE&(RWS==MADR)) | (MBFL&~MBWE&(RWS==MBDR));
  end
  always @(STRR or STRW or MAWE or MDRDY or ALUWP or RWES or CCOPS or DLSCT or
	   IVFT or MBFL or DALUCT) begin
    STALL = STRR | STRW | (MAWE&MDRDY&~ALUWP&(RWES|CCOPS[0]));
    STALL = STALL | (DLSCT[2]&MBFL);
    STALL = STALL | ((DALUCT[4:2]==5)&MBFL);
  end
  always @(RWES or STRW or MAWE or MDRDY or MADR or IRQST) begin
    RWE = RWES & ~STRW & ~(MAWE&MDRDY);
    RWE = RWE | (MAWE&MDRDY&~((MADR==5'h1f)&(IRQST==3)));
  end 
  always @(STRW or MAWE or MDRDY or ALUWP or CCOPS)
    CCOP = ~STRW & ~(MAWE&MDRDY&~ALUWP) ? CCOPS : 2'd0;
  always @(LCTDS)
    LCTD = LCTDS; 
  always @(MAWE or MDRDY or ALUWP or MADR or RWS)
    RBWA = MAWE & MDRDY & ~ALUWP ? MADR : RWS;
  always @(IRQST or MAWE or MDRDY or MADR)
    SOIR = (IRQST==3) & MAWE & MDRDY & (MADR==5'h1f);
  always @(IBFL or MAST or DRRE or RWES or CCOPS or LCTDS or STALL or DIWV or
           DRWE or DLSCT or MBFL) begin
    RFST = (IBFL==0) & ~MAST & ~DRRE & ~(RWES|CCOPS[0]|LCTDS);
    RFST = RFST & ~STALL & ~DIWV & (DRWE==0) & ~DLSCT[2] & ~MBFL;
  end
  always @(SHCT or DALUCT or IPH or IW or IBFL or RBDI) begin
    NCSHCT = SHCT;
    if((SHCT!=0) & (DALUCT[4:3]==3) & (IPH!=3))
      NCSHCT = NCSHCT - 1'b1;
    if((IW[19:18]==0) & (IW[3:0]==6) & (IBFL!=0) & (IPH==0))
      NCSHCT = IW[15:12];
    if((IW[19:18]==0) & (IW[3:0]==2) & (IPH==1))
      NCSHCT = RBDI[3:0];
  end
  always @(DRWE or RWES or RWS or DCCOPC or CCOPS)
    CCPO = (DRWE==5'h10) | (RWES&(RWS==5'h10)) | (DCCOPC!=0) | (CCOPS!=0);
  always @(DRWE or RWES or RWS)
    LCPO = (DRWE==5'h12) | (RWES&(RWS==5'h12));
  always @(DRWE or RWES or RWS or MBFL or MBDR or DLSCT or DRRS or MAWE or
           MADR)
    TAPO = (DRWE==5'h1e) | (RWES&(RWS==5'h1e));
  always @(DRWE or RWES or RWS or MBFL or MBDR or DLSCT or DRRS or MAWE or
           MADR) begin
    SAPO = (DRWE==5'h1c) | (RWES&(RWS==5'h1c)) | (MBFL&(MBDR==5'h1c));
    SAPO = SAPO | (((DLSCT[2:1]==2)&(DRRS==5'h1c))|(MAWE&(MADR==5'h1c)));
  end
  always @(IW or IBFL or FLIV or IRQST or PCCT or NSIF or IIV or TAPO or IPH or
           SAPO or TAPO or CCPO or LCPO or DCCOPC or CCOPS or RBDI or SHCT or
           LCTST or BCDT or SOIR or DSOIR or STALL or RWEC or RRSC or CSF)
  begin
    FLIV   = 0;
    FEXR   = 1;
    IACT   = 0;
    FLCT   = 0;
    CCSCT  = 0;
    STPCT  = 0;
    PCCT   = 0;
    NCNSIF = 0;
    SQFFI  = 0;
    RRSC   = 0;
    RWEC   = 0;
    CCOPC  = 0;
    DBGC   = 0;
    AS0CT  = 0;
    AS1CT  = 0;
    ACOCT  = 0;
    ALUCT  = 0;
    LSCT   = 0;
    LSASL  = 0;
    LDSTM  = 0;
    NIPH   = IPH;
    if(IBFL) begin
      if(~IW[1]) begin
	if(IW[19] & IW[0]) begin                        
          FLIV = 1;
          if(~IW[18])
            idec_jpsr_ia16;
          else if(IW[3])
            if((IW[6:4]==7) & IW[2])
              idec_brlc;
            else
              idec_cdbranch;
          else if(~IW[2])
            idec_bral;
          else if(IW[7]) begin
            FLCT  = {1'b1,IW[5:4]};
	    SQFFI = 1;
	    FEXR  = FEXR & ~SAPO & (IRQST==0);
	  end
          else
            case(IW[6:4])
              0,
              1: idec_jump_ta;
              2,
              3: idec_return;
              4: idec_stop;
              6: PCCT[0] = 1;
              7: PCCT[1] = 1;
	      default: ;
            endcase
        end
        else                                            
          idec_ldst;
      end
      else begin                                        
	if(IW[19]) begin
          if(IW[3:2] == 0)
            idec_con8;
	  else
            idec_con10;
	end
	else if(~IW[3])
          idec_shfbit;
        else if(~IW[2])
          idec_triarn;
        else if(~IW[18])
          idec_mult;
        else if(IW[16]) begin
          if(IW[15])
            idec_compreg;
          else
            idec_mtfsrdp;
	end
        else
          idec_dual;
      end
    end
    NCNSIF = (NCNSIF&FLIV) | DSOIR | PCCT[1] | (NSIF&~IIV);
    CCSCT[0] = SOIR;
    if(STALL & FLIV & (RWEC[4]|RRSC[5])) begin
      FEXR   = 0;
      NCNSIF = 0;
    end
  end
  always @(STALL)
    AEXRD = ~STALL;
  always @(NIPH or STALL or FEXR or DLDSTM)
    RFNXI = (NIPH==0) & ~STALL & FEXR & ~DLDSTM;
  always @(IRST or QDIW or FLIV or SQFFI or FEXR or NCNSIF or NSIF or IRQST or
	   LSTRQ or LDBRQ) begin
    IFT = ~IRST & (QDIW<3) & (IRQST==0);
    if(FLIV & ~SQFFI)
      IFT = FEXR & NCNSIF & ~NSIF;
    if(LSTRQ | LDBRQ)
      IFT = 0;
  end
  always @(IW or PC) begin
    BOFS = {6'h0,IW[17:8]};
    if(IW[3]) begin
      if(IW[17])
        BOFS[15:10] = 6'h3f;
    end
    else
      BOFS[13:10] = IW[7:4];
    if(BOFS[13])
      BOFS[15:14] = 2'd3;
    BTIA = PC + BOFS;
  end
  always @(IBFL or RFNXI)
    IWCS = (IBFL!=0) & RFNXI;
  always @(IRQST or LSTRQ or LDBRQ or STPCT or IRQ or CSF or FLCT or RFST or
	   MBFL or MAWE or MDRDY or MADR) begin
    NCIRQST = IRQST;
    case(IRQST)
      0: if(~LSTRQ & ~LDBRQ & ~STPCT)
           if(IRQ & (CSF[1:0]==2) & ~(FLCT[2]&FLCT[0]))
             NCIRQST = 1;
      1: begin
           if(RFST)
             NCIRQST = 2;
           if(STPCT)
             NCIRQST = 0;
         end
      2: NCIRQST = MBFL ? 2'd1 : 2'd3;
      3: if(MAWE & MDRDY & (MADR==5'h1f))
           NCIRQST = 0;
    endcase
  end
  always @(NCNSIF or NSIF)
    IFNS = NCNSIF & ~NSIF;
  always @(IACT or IW or BTIA or RBDI or MAST or MAIA) begin
    IA = 0;
    if(IACT) begin
      if(IACT[0])
        IA = IW[17:2]; 
      if(IACT[1])
        IA = BTIA;
      if(IACT[2])
        IA = RBDI;
    end
    else
      IA = MAST ? MAIA + 1'b1 : MAIA;
  end
  always @(MAST or IIV or NCNSIF or NSIF or IFT or STPCT or FEXR or SOIR or
           FLIV) begin
    NCMAST = MAST;
    if(~MAST | IIV | (NCNSIF&~NSIF) | (FLIV&~FEXR))
      NCMAST = IFT;
    if(STPCT & FEXR)
      NCMAST = 0;
  end
  always @(CSF or CCSCT or SOIR or FLCT or FEXR or STALL) begin
    NCCSF = CSF;
    if(CCSCT[1])
      NCCSF[0] = 0;
    if(SOIR)
      NCCSF[0] = 1;
    if(FLCT[2] & FEXR & ~STALL)
      case(FLCT[1:0])
        0: NCCSF[1] = 1;
        1: NCCSF[1] = 0;
        2: NCCSF[1] = CSF[2];
        3: begin
             NCCSF[2] = CSF[1];
             NCCSF[1] = 0;
           end
      endcase
  end
  always @(IBFL or IIV or IWCS or NCNSIF or NSIF or SOIR or STPCT or FEXR or
	   LINJI or FLCT or IRQST) begin
    NCIBFL = IBFL + IIV - IWCS;
    if((NCNSIF&~NSIF&FEXR) | SOIR | (STPCT&FEXR) |
        (FLCT[2]&FLCT[0]&(IRQST==1)))
      NCIBFL = 0;
    if(LINJI)
      NCIBFL = 1;
  end
  always @(STPCT or FEXR or LDBRQ or LSTRQ)
    NCLDBRQ = (STPCT&FEXR) | (LDBRQ&~LSTRQ);
  always @(DS0DI) begin
    LCTLD[0] = (DS0DI[15:2]==0) & (DS0DI[1:0]!=3);
    LCTLD[1] = LCTLD[0] & ~DS0DI[1];
  end
  always @(IW) begin
    RFCM = 20'hfffff;
         if(IW[ 4]) begin 
                      SMRS = IW[2] ? 5'h0b : 5'h0d;
                      LMRS = IW[2] ? 5'h1c : 5'h00;
                      RFCM[4] = 0;
                    end
    else if(IW[ 5]) begin
                      SMRS = IW[2] ? 5'h0a : 5'h0c;
                      LMRS = 5'h02;
                      RFCM[5] = 0;
                    end
    else if(IW[ 6]) begin
                      SMRS = IW[2] ? 5'h09 : 5'h01;
                      LMRS = 5'h03;
                      RFCM[6] = 0;
                    end
    else if(IW[ 7]) begin SMRS = 5'h07; LMRS = 5'h04; RFCM[ 7] = 0; end
    else if(IW[12]) begin SMRS = 5'h06; LMRS = 5'h05; RFCM[12] = 0; end
    else if(IW[13]) begin SMRS = 5'h05; LMRS = 5'h06; RFCM[13] = 0; end
    else if(IW[14]) begin SMRS = 5'h04; LMRS = 5'h07; RFCM[14] = 0; end
    else if(IW[15]) begin
                      SMRS = 5'h03;
                      LMRS = IW[2] ? 5'h09 : 5'h01;
                      RFCM[15] = 0;
                    end
    else if(IW[16]) begin
                      SMRS = 5'h02;
                      LMRS = IW[2] ? 5'h0a : 5'h0c;
                      RFCM[16] = 0;
                    end
    else            begin 
                      SMRS = IW[2] ? 5'h1c : 5'h00;
                      LMRS = IW[2] ? 5'h0b : 5'h0d;
                    end
  end
  always @(RRSC or CSF or IW or SMRS or LMRS or IRQST or IBFL or STALL or DRRS)
  begin
    RRS = 0;
    if(RRSC[5])
      RRS = RRSC[4:0];
    else begin
      if(RRSC[0])
        RRS[3:0] = IW[15:12];
      if(RRSC[1])
        RRS[3:0] = IW[11:8];
      if(RRSC[2])
        RRS[3:0] = IW[7:4];
      if(RRSC[3])
        RRS[3:0] = {1'b1,IW[10:8]}; 
      if(RRSC[4])
        RRS = IW[11] ? SMRS : LMRS;
    end
    if((IRQST==1) & (IBFL==0))
      RRS = 5'h11;
    if(STALL)
      RRS = DRRS;
  end
  always @(RWS or AEXRD or DRWE or DIW or SOIR) begin
    NCRWS = RWS;
    if(AEXRD)
      if(DRWE[4])
        NCRWS = {1'b1,DRWE[3:0]};
      else begin
        if(DRWE[0])
          NCRWS = {1'b0,DIW[7:4]};
        if(DRWE[1])
          NCRWS = {2'h1,DIW[10:8]}; 
      end
    if(SOIR)
      NCRWS = 5'h1d;
  end
  always @(RWES or AEXRD or DRWE or SOIR or MAWE or ALUWP or MDRDY or STRW)
  begin
    NCRWES = RWES;
    if(AEXRD) begin
      NCRWES = DRWE != 0;
      if(SOIR)
        NCRWES = 1;
    end
    else if(~STRW & ~(MAWE&MDRDY))
      NCRWES = 0;
  end
  always @(posedge CLK) begin
    if(DBGWE)
      DBGO <= DS0DI;
    if(PCCT[0])
      DBGO <= MAIA;
    if(AEXRD) begin
      CCOPS <= DCCOPC;
      DBGWE <= DDBGC;
      LCTDS <= (DRWE==5'h12) & DRRE & (DRRS==5'h12);
    end
    else begin
      if((~(MAWE&MDRDY)|(ALUWP&MDRDY)) & ~STRW) begin
        CCOPS <= 0;
        LCTDS <= 0;
      end
      DBGWE <= 0;
    end
    if(~STALL) begin
      DIW    <= IW;
      DIWV   <= IPH == 0;
      DACOCT <= ACOCT;
      DALUCT <= ALUCT;
      DLSCT  <= LSCT;
      DDBGC  <= DBGC;
      DLSASL <= LSASL;
      DAS0CT <= AS0CT;
      DAS1CT <= AS1CT;
      DCCOPC <= CCOPC;
      DLDSTM <= LDSTM;
      DRRS   <= RRS;
      DRWE   <= RWEC;
      DRRE   <= (RRSC!=0) & (~LSCT[2]|LSCT[1]);
    end
    if(~STALL & (IBFL==0)) begin
      DALUCT <= 0;
      DLSCT  <= 0;
      DAS0CT <= 2;
      DAS1CT <= 0;
      DRRE   <= 0;
      DDBGC  <= 0;
      DLDSTM <= 0;
    end
    if((~STALL&(IBFL==0)) | FLIV)
      DIWV <= 0;
    if(~STALL)
      IPH <= NIPH;
    if(NCIBFL == 0)
      IPH <= 0;
    if(CCSCT[0])
      CCS <= CC;
    if(CCSCT[1])
      CC <= CCS;
    else
      if(CCOP[0])
        CC <= CFI;
    if((IRQST==0) & (NCIRQST==1)) begin
      IACK <= 1;
      LIRN <= IRN;
    end
    else
      IACK <= 0;
    if(RWE) begin
      if(RBWA == 5'h12)
        if(LCTD)
          LCTST <= ~LCTLD[0];
        else
          LCTST <= ~LCTLD[1];
      if(RBWA == 5'h10)
        CC <= DS0DI[3:0];
    end
    if((RRSC[5:4]==1) & ~STALL)
      IW <= IW & RFCM;
    if(IWCS & IBFL[1]) begin
      IW  <= IB0;
      IB0 <= IB1;
    end
    if(IIV)
      if((IBFL==0) | ((IBFL==1)&IWCS))
        IW <= II;
      else if((IBFL==1) | IWCS)
        IB0 <= II;
      else
        IB1 <= II;
    if(LINJI)
      IW <= DBGIR;
    if(!STPCT)
      if(NSIF & IIV)
        PC <= MAIA;
      else if(IWCS & ~INDB)
        PC <= PC + 1'b1;
    if(STPCT)
      MAIA <= PC;
    else if(SOIR)
      MAIA = MDI[13:0];
    else if(IIV | (NCNSIF&~NSIF) | (LSTRQ&FLIV&~SQFFI))
      MAIA <= IA;
    if(PCCT[1])
      MAIA <= DBGIR[13:0];
    if(STPD) begin
      DBGIR <= DBGI;
      LINJI <= INJI;
    end
    INDB <= (INDB&LSTRQ) | STPD;
    STPD  <= (LSTRQ|LDBRQ) & RFST & (IRQST==0);
    LSTRQ <= STRQ & (IRQST==0) & (NCIRQST==0);
    DSOIR <= SOIR;
    CSF   <= NCCSF;
    MAST  <= NCMAST;
    IBFL  <= NCIBFL;
    NSIF  <= NCNSIF;
    RWS   <= NCRWS;
    RWES  <= NCRWES;
    IRQST <= NCIRQST;
    LDBRQ <= NCLDBRQ;
    if(~STALL)
      SHCT <= NCSHCT;
    if(IRST) begin
      IPH     <= 0;
      STPD    <= 0;
      LSTRQ   <= 0;
      LDBRQ   <= 0;
      INDB    <= 0;
      LINJI   <= 0;
      CC      <= 0;
      CSF     <= 1;
      CCS     <= 0;
      MAST    <= 0;
      NSIF    <= 1;
      DSOIR   <= 0;
      LCTST   <= 0;
      IRQST   <= 0;
      IACK    <= 0;
      IBFL    <= 0;
      MAIA    <= {IRN,12'h0};
      PC      <= {IRN,12'h0};
      DIWV    <= 0;
      DALUCT  <= 0;
      DLSCT   <= 0;
      DLDSTM  <= 0;
      RWES    <= 0;
      CCOPS   <= 0;
      DDBGC   <= 0;
      DBGWE   <= 0;
      LCTDS   <= 0;
      DRRE    <= 0;
      DRWE    <= 0;
      SHCT    <= 0;
    end
    if(RST)
      RSYN <= 3;
    else if(RSYN[1])
      RSYN[1] <= 0;
    else
      RSYN <= 0;
  end
endmodule
module sf20buDpu(CLK,RST,MDI,MDRDY,RSD,RFE0,RFE1,IW,DBGI,CC,EXRD,ALUWP,S0CT,
                 S1CT,COCT,DCCOPC,CCOP,DRRS,RWS,RWE,LSCT,LSASL,IVFT,LIRN,PC,
                 IRQST,OPCT,DO,CFO,OR0,MDA,MDO,MDBS,MDWE,MBFL,MBDR,MADR,MAWE);
  input         CLK;
  input         RST;
  input  [15:0] MDI;
  input         MDRDY;
  input  [15:0] RSD;
  input  [ 1:0] RFE0;
  input  [ 1:0] RFE1;
  input  [19:0] IW;
  input  [19:0] DBGI;
  input  [ 3:0] CC;
  input         EXRD;
  input         ALUWP;
  input  [ 2:0] S0CT;
  input  [ 2:0] S1CT;
  input  [ 4:0] COCT;
  input  [ 1:0] DCCOPC;
  input  [ 1:0] CCOP;
  input  [ 4:0] DRRS;
  input  [ 4:0] RWS;
  input         RWE;
  input  [ 2:0] LSCT;
  input         LSASL;
  input         IVFT;
  input  [ 3:0] LIRN;
  input  [15:0] PC;
  input  [ 1:0] IRQST;
  input  [ 4:0] OPCT;
  output [15:0] DO;
  output [ 3:0] CFO;
  output [15:0] OR0;
  output [15:0] MDA;
  output [15:0] MDO;
  output [ 1:0] MDBS;
  output        MDWE;
  output        MBFL;
  output [ 4:0] MBDR;
  output [ 4:0] MADR;
  output        MAWE;
  reg    [15:0] DO;
  reg    [ 3:0] CFO;
  reg    [15:0] OR0;
  reg    [15:0] MDA;
  reg    [15:0] MDO;
  reg    [ 1:0] MDBS;
  reg           MDWE;
  reg           MBFL;
  reg    [ 4:0] MBDR;
  reg    [ 4:0] MADR;
  reg           MAWE;
  reg    [ 4:0] OPR;
  reg    [15:0] OR1;
  reg           CRIN;
  reg    [15:0] BMSK;
  reg    [15:0] PCBF;
  reg           PCBV;
  reg signed [16:0] MLS0;
  reg signed [16:0] MLS1;
  reg    [ 1:0] MBOC;
  reg           MAA;
  reg    [ 2:0] MAOC;
  reg    [15:0] FMDO;
  reg    [15:0] CON;
  reg    [15:0] SADD;
  reg    [15:0] SHFD;
  reg    [15:0] LBS0;
  reg    [15:0] LBTD;
  reg    [15:0] ALUD;
  reg    [ 3:0] SACC;
  reg    [ 3:0] LBCC;
  reg           MB0;
  reg           MB1;
  reg           MBA;
  reg    [ 3:0] BTIX;
  reg    [15:0] SD0;
  reg    [15:0] SD1;
  reg    [31:0] MUL;
  reg    [15:0] MULO;
  reg           NCMBFL;
  reg    [ 2:0] NCMAOC;
  reg    [15:0] NMDA;
  always @(MDI or MAOC or MAA or MDI) begin
    FMDO = MDI;
    if(~MAOC[0])
      if(MAA)
        FMDO = {8'h0,MDI[15:8]};
      else
        FMDO[15:8] = 0;
  end
  always @(COCT or IW) begin
    CON = 0;
    if(COCT[3]) begin
      CON[7:0] = {IW[12:11],IW[18:13]}; 
      if(COCT[0])
        CON[9:8] = IW[10:9]; 
      if((COCT[1]&IW[8]) | (COCT[2]&COCT[0]&IW[10]))
        CON[15:10] = 6'h3f; 
      if(COCT[2] & ~COCT[0] & IW[12])
        CON[15:8] = 8'hff; 
    end
    else if(COCT[2])
      if(COCT[0]) begin
        CON[15:1] = 15'h7fff;
	if(~COCT[1])
          CON[0] = 1;
      end
      else
        CON[1:0] = COCT[1] ? 2'd2 : 2'd1; 
    if(COCT[4])
      CON[15:8] = {IW[12:11],IW[18:13]}; 
  end
  always @(S1CT or IW or RSD)
    BTIX = S1CT[0] ? IW[15:12] : RSD[3:0];
  always @(OR1 or OR0 or CRIN or OPR or CCOP or CC) begin
    SADD = OPR[0] ? OR1 + OR0 + CRIN : OR1 - OR0 - CRIN;
    MB0 = OR0[15];
    MB1 = OR1[15];
    MBA = SADD[15];
    if(OPR[0]) begin
      SACC[0] = (MB1&MB0) | (MB1&~MBA) | (MB0&~MBA);
      SACC[1] = ((MB1&MB0&~MBA) | (~MB1&~MB0&MBA));
    end
    else begin
      SACC[0] = (~MB1&MB0) | (~MB1&MBA)|(MB0& MBA);
      SACC[1] = ((MB1&~MB0&~MBA) | (~MB1&MB0&MBA));
    end
    SACC[2] = SADD == 0;
    SACC[3] = SADD[15];
    if(CCOP[1] & ~CC[2])
      SACC[2] = 0;
  end
  always @(OPR or OR1) begin
    SHFD = OPR[0] ? {1'b0,OR1[15:1]} : {OR1[14:0],1'b0};
    if(OPR[1:0] == 2)
      SHFD[0] = OR1[15];
    if(OPR[1:0] == 3)
      SHFD[15] = OR1[15];
  end
  always @(OPR or OR0 or OR1 or CC or BMSK) begin
    LBS0 = OPR[2] ? BMSK : OR0;
    case(OPR[1:0])
      0: LBTD = OR1 & ~LBS0;              
      1: LBTD = OR1 &  LBS0;              
      2: LBTD = OR1 |  LBS0;              
      3: LBTD = OR1 ^  LBS0;              
    endcase
    LBCC = 4'h0;
    if(~OPR[2]) begin
      LBCC[0] =           LBTD[ 0] ^ LBTD[ 1] ^ LBTD[ 2] ^ LBTD[ 3];
      LBCC[0] = LBCC[0] ^ LBTD[ 4] ^ LBTD[ 5] ^ LBTD[ 6] ^ LBTD[ 7];
      LBCC[0] = LBCC[0] ^ LBTD[ 8] ^ LBTD[ 9] ^ LBTD[10] ^ LBTD[11];
      LBCC[0] = LBCC[0] ^ LBTD[12] ^ LBTD[13] ^ LBTD[14] ^ LBTD[15];
      LBCC[1] = CC[1]  ^ LBCC[0];
    end
    LBCC[2] = LBTD == 0;
    LBCC[3] = LBTD[15];
  end
  always @(MLS0 or MLS1 or OPR) begin
    MUL  = MLS0 * MLS1;
    MULO = OPR[1] ? MUL[31:16] : MUL[15:0];
  end
  always @(OR0 or OPR or SHFD or MULO or SADD or SACC or LBTD or LBCC or
	   MAWE or MDRDY or ALUWP or FMDO) begin
    ALUD = OR0;
    CFO  = 0;
    if(OPR[4:3] == 3)                              
      ALUD = SHFD;
    else if(OPR[4:3] == 1)                         
      ALUD = MULO;
    else if(OPR[4:3] == 0)
      case(OPR[2:0])
        0,
        1: begin                                   
             ALUD = SADD;
             CFO  = SACC;
           end
        2: if(OR0[15])                             
             ALUD = SADD;
        3: ALUD = OR0[15] ? 16'hffff : 16'h0;      
        6: if(OR0[15]) ALUD = 16'h0000;            
           else if(OR0[14]) ALUD = 16'h0001;
           else if(OR0[13]) ALUD = 16'h0002;
           else if(OR0[12]) ALUD = 16'h0003;
           else if(OR0[11]) ALUD = 16'h0004;
           else if(OR0[10]) ALUD = 16'h0005;
           else if(OR0[ 9]) ALUD = 16'h0006;
           else if(OR0[ 8]) ALUD = 16'h0007;
           else if(OR0[ 7]) ALUD = 16'h0008;
           else if(OR0[ 6]) ALUD = 16'h0009;
           else if(OR0[ 5]) ALUD = 16'h000a;
           else if(OR0[ 4]) ALUD = 16'h000b;
           else if(OR0[ 3]) ALUD = 16'h000c;
           else if(OR0[ 2]) ALUD = 16'h000d;
           else if(OR0[ 1]) ALUD = 16'h000e;
           else if(OR0[ 0]) ALUD = 16'h000f;
           else ALUD = 16'h0010;
        7: ALUD[15:8] = OR0[7] ? 8'hff : 8'h00;    
        default: ;
      endcase
    else begin                                     
      ALUD = LBTD;
      CFO  = LBCC;
    end
    DO = MAWE & MDRDY & ~ALUWP ? FMDO : ALUD;
  end
  always @(RFE0 or ALUD or RSD or FMDO)
    if(RFE0[0])
      SD0 = ALUD;
    else if(RFE0 == 0)
      SD0 = RSD;
    else
      SD0 = FMDO;
  always @(RFE1 or ALUD or RSD or FMDO)
    if(RFE1[0])
      SD1 = ALUD;
    else if(RFE1 == 0)
      SD1 = RSD;
    else
      SD1 = FMDO;
  always @(MBOC or BMSK or MDA) begin
    MDO = 0;
    if(MBOC[0])
      MDO = BMSK;
    else
      MDO = MDA[0] ? {BMSK[7:0],8'h0} : {8'h0,BMSK[7:0]};
  end
  always @(LSASL or OR1 or ALUD)
    NMDA = LSASL ? OR1 : ALUD;
  always @(MBFL or MAOC or MDRDY or LSCT or IVFT or EXRD) begin
    NCMBFL = MBFL;
    if(~MAOC[2] | MDRDY)
      NCMBFL = 0;
    if((LSCT[2]|IVFT) & ~MBFL & EXRD)
      NCMBFL = 1;
  end
  always @(MAOC or MDRDY or MBFL or MBOC) begin
    NCMAOC = MAOC;
    if(MAOC[2] & MDRDY)
      NCMAOC[2] = 0;
    if(MBFL & (~MAOC[2]|MDRDY))
      NCMAOC = {1'b1,MBOC};
  end
  always @(posedge CLK) begin
    if(EXRD) begin
      OPR <= OPCT;
      if(DCCOPC[1]) begin
	CRIN <= (~CCOP[0]&CC[0]) | (CCOP[0]&SACC[0]);
	if(RWE & (RWS==5'h10) & ~MAWE)
	  CRIN <= OR0[0];
      end
      else
	CRIN <= 0;
    end
    if((IRQST==2) & ~MBFL)
      OPR <= 4;
    if(EXRD & (OPCT[4:3]==1)) begin
      if(S0CT[0])
        MLS0 <= {SD0[15]&OPCT[0],SD0};
      if(S0CT[1])
        MLS0 <= {1'b0,CON};
      if(S1CT[0])
        MLS1 <= {SD1[15]&OPCT[0],SD1};
    end
    if((OPCT[4:2]==5) & (IW[2]|~S1CT) & EXRD & ~MBFL) begin
      BMSK <= 0;
      case(BTIX)
        4'h0: BMSK[ 0] <= 1;
        4'h1: BMSK[ 1] <= 1;
        4'h2: BMSK[ 2] <= 1;
        4'h3: BMSK[ 3] <= 1;
        4'h4: BMSK[ 4] <= 1;
        4'h5: BMSK[ 5] <= 1;
        4'h6: BMSK[ 6] <= 1;
        4'h7: BMSK[ 7] <= 1;
        4'h8: BMSK[ 8] <= 1;
        4'h9: BMSK[ 9] <= 1;
        4'ha: BMSK[10] <= 1;
        4'hb: BMSK[11] <= 1;
        4'hc: BMSK[12] <= 1;
        4'hd: BMSK[13] <= 1;
        4'he: BMSK[14] <= 1;
        4'hf: BMSK[15] <= 1;
      endcase
    end
    if(MAOC[2] & MDRDY)
      MAWE <= 0;
    if(MBFL & (~MAOC[2]|MDRDY)) begin
      MADR <= MBDR;
      MAWE <= ~MDWE;
      MAA  <= MDA[0];
    end
    if(~MAOC[2] | MDRDY) begin
      MDBS <= 0;
      MDWE <= 0;
    end
    if(IVFT & EXRD) begin
      MBDR <= 5'h1f;
      MBOC <= 2;
      MDBS <= 3;
      MDA  <= {RSD[15:5],LIRN,1'b0};
    end
    else if(LSCT[2] & ~MBFL & EXRD) begin
      BMSK <= SD0;
      MBDR <= DRRS;
      MBOC <= LSCT[1:0];
      MDA  <= NMDA;
      MDWE <= LSCT[1];
      MDBS <= LSCT[0] ? 2'd3 : (NMDA[0] ? 2'd2 : 2'd1);
    end
    if(EXRD)
      if(S0CT[0]) begin
        OR0 <= SD0;
        if((IW[19:17]==4) & (IW[2:0]==4))
          OR0 <= {SD0[14:0],1'b0};
      end
      else if(S0CT[1])
        OR0 <= CON;
      else if(S0CT[2])
        OR0 <= DBGI[15:0];
    if(EXRD) begin
      if(S1CT[0] | (~S1CT[0]&RFE1[0]))
        OR1 <= SD1;
      else if(S1CT[1])
        OR1 <= CON;
      if(((IRQST==2)&~MBFL) | (~PCBV&S1CT[2]))
        OR1 <= PC;
      if(PCBV) begin
        OR1  <= PCBF;
        PCBV <= 0;
      end
    end
    if(~EXRD & S1CT[2] & ~PCBV) begin
      PCBF <= PC;
      PCBV <= 1;
    end
    MBFL <= NCMBFL;
    MAOC <= NCMAOC;
    if(RST) begin
      OPR   <= 0;
      MDBS  <= 0;
      MDWE  <= 0;
      MBFL  <= 0;
      MBOC  <= 0;
      MAOC  <= 0;
      MAWE  <= 0;
      PCBV  <= 0;
    end
  end
endmodule
module sf20buRf(CLK,RST,DRRS,CC,CCS,CSF,RBDO,RBDI,RBWA,RBWE,DI,RP,RDTW);
  input         CLK;
  input         RST;
  input  [ 4:0] DRRS;
  input  [ 3:0] CC;
  input  [ 3:0] CCS;
  input  [ 2:0] CSF;
  input  [15:0] RBDO;
  input  [15:0] RBDI;
  input  [ 4:0] RBWA;
  input         RBWE;
  input  [15:0] DI;
  output [15:0] RP;
  output [15:0] RDTW;
  reg    [15:0] RP;
  reg    [15:0] RDTW;
  reg           LWDV;
  reg    [ 4:0] LWDR;
  reg    [15:0] LWDB;
  always @(LWDV or DRRS or LWDR or LWDB or RBDO or CC or CSF) begin
    if(LWDV & (DRRS==LWDR))
      RP = LWDB;
    else
      RP = RBDO;
    if(DRRS == 5'h10)
      RP = {12'h0,CC};
    if(DRRS == 5'h11)
      RP[4:0] = {2'h0,CSF};
    if(DRRS == 5'h1f)
      RP = 16'h1417;
    if((DRRS[4:2]==5) & RP[9])
      RP[15:10] = 6'h3f;
  end
  always @(DI or RBWA) begin
    RDTW = DI;
    if((RBWA==5'h12) | (RBWA[4:2]==5))
      RDTW[15:10] = 0;
  end
  always @(posedge CLK) begin
    if(RBWE) begin
      LWDB <= RBDI;
      LWDR <= RBWA;
      LWDV <= 1;
    end
    if(RST)
      LWDV <= 0;
  end
endmodule
module sf20buRfb(CLK,RAD,DI,WAD,WE,DO);
  input         CLK;
  input  [ 4:0] RAD;
  input  [15:0] DI;
  input  [ 4:0] WAD;
  input         WE;
  output [15:0] DO;
  reg    [15:0] DO;
  reg    [ 4:0] DRAD;
  reg    [15:0] R0;
  reg    [15:0] R1;
  reg    [15:0] R2;
  reg    [15:0] R3;
  reg    [15:0] R4;
  reg    [15:0] R5;
  reg    [15:0] R6;
  reg    [15:0] R7;
  reg    [15:0] R8;
  reg    [15:0] R9;
  reg    [15:0] RA;
  reg    [15:0] RB;
  reg    [15:0] RC;
  reg    [15:0] RD;
  reg    [15:0] RE;
  reg    [15:0] RF;
  reg    [15:0] D0;
  reg    [15:0] IV;
  reg    [15:0] LC;
  reg    [15:0] D1;
  reg    [15:0] U0;
  reg    [15:0] U1;
  reg    [15:0] U2;
  reg    [15:0] U3;
  reg    [15:0] AE0;
  reg    [15:0] AE1;
  reg    [15:0] AE2;
  reg    [15:0] AE3;
  reg    [15:0] SA;
  reg    [15:0] IA;
  reg    [15:0] TA;
  reg    [15:0] D2;
  always @(DRAD or R0 or R1 or R2 or R3 or R4 or R5 or R6 or R7 or R8 or R9 or
           RA or RB or RC or RD or RE or RF or D0 or IV or LC or D1 or U0 or
           U1 or U2 or U3 or AE0 or AE1 or AE2 or AE3 or SA or IA or TA or D2)
    case(DRAD)
      5'h00: DO = R0;
      5'h01: DO = R1;
      5'h02: DO = R2;
      5'h03: DO = R3;
      5'h04: DO = R4;
      5'h05: DO = R5;
      5'h06: DO = R6;
      5'h07: DO = R7;
      5'h08: DO = R8;
      5'h09: DO = R9;
      5'h0a: DO = RA;
      5'h0b: DO = RB;
      5'h0c: DO = RC;
      5'h0d: DO = RD;
      5'h0e: DO = RE;
      5'h0f: DO = RF;
      5'h10: DO = D0;
      5'h11: DO = IV;
      5'h12: DO = LC;
      5'h13: DO = D1;
      5'h14: DO = U0;
      5'h15: DO = U1;
      5'h16: DO = U2;
      5'h17: DO = U3;
      5'h18: DO = AE0;
      5'h19: DO = AE1;
      5'h1a: DO = AE2;
      5'h1b: DO = AE3;
      5'h1c: DO = SA;
      5'h1d: DO = IA;
      5'h1e: DO = TA;
      5'h1f: DO = D2;
    endcase
  always @(posedge CLK) begin
    if(WE)
      case(WAD)
        5'h00: R0 <= DI;
        5'h01: R1 <= DI;
        5'h02: R2 <= DI;
        5'h03: R3 <= DI;
        5'h04: R4 <= DI;
        5'h05: R5 <= DI;
        5'h06: R6 <= DI;
        5'h07: R7 <= DI;
        5'h08: R8 <= DI;
        5'h09: R9 <= DI;
        5'h0a: RA <= DI;
        5'h0b: RB <= DI;
        5'h0c: RC <= DI;
        5'h0d: RD <= DI;
        5'h0e: RE <= DI;
        5'h0f: RF <= DI;
        5'h10: D0 <= DI;
        5'h11: IV <= DI;
        5'h12: LC <= DI;
        5'h13: D1 <= DI;
        5'h14: U0 <= DI;
        5'h15: U1 <= DI;
        5'h16: U2 <= DI;
        5'h17: U3 <= DI;
        5'h18: AE0 <= DI;
        5'h19: AE1 <= DI;
        5'h1a: AE2 <= DI;
        5'h1b: AE3 <= DI;
        5'h1c: SA <= DI;
        5'h1d: IA <= DI;
        5'h1e: TA <= DI;
        5'h1f: D2 <= DI;
      endcase
    DRAD <= RAD;
  end
endmodule
